package Pack;

import lejos.hardware.Brick;
import lejos.hardware.BrickFinder;
import lejos.hardware.Button;
//import lejos.hardware.Sound;
import lejos.hardware.lcd.LCD;
import lejos.hardware.motor.Motor;
import lejos.hardware.port.Port;
import lejos.hardware.sensor.EV3ColorSensor;
import lejos.robotics.navigation.MovePilot;
import lejos.utility.Delay;


/*
 * ナップザック問題を解いて、ベルトコンベアーの上に載っている荷物を取捨選択する
 * 今回、荷物の数は決まっているものとする
 */
public class EV3Sorter {

	// 荷物の数は4とする
	public static byte NUMBER_GOODS = 5;

	// 重さの制限
	static int WEIGHT_LIMIT = 30;

	static int[] goodsWeight = new int[NUMBER_GOODS];
	static int[] goodsValue = new int[NUMBER_GOODS];

	// 色と重さの変換表
	static int weightColor2Weight[] = {
			5, // RED : 5 kg
			-1,//Green : nothing
			10,// BLUE : 10 kg
			15 // YELLOW : 15 kg
	};

	// 色と価値の変換表
	static int valueColor2Value[] = {
			100, // RED : 100 yen
			-1,  // Green : nothing
			300, // BLUE : 300 yen
			500  // YELLOW : 500 yen
	};

	public static byte NUMBER_ACTIONS = 2;
	public static byte NUMBER_STATES = 4;

	static MovePilot pilot;
	static Sensor color;

	byte motorCommand = 0;

	private static EV3ColorSensor sensor3;

//	@SuppressWarnings("deprecation")

	public static void main(String[] args) {

		Brick brick1 = BrickFinder.getDefault();
		Port s3 = (Port)brick1.getPort("S3");
		sensor3 = new EV3ColorSensor(s3);
		color = new Sensor(sensor3.getColorIDMode());

		byte env = -1;

		//初期状態として、アーム(ハンマー)はコンベアー上から退避する→しない

		/* STEP 1
		 *  荷物の全容を把握する
		 *  荷物はあらかじめセンサーの手前に整列しておかれていると仮定
		 */
		
		Delay.msDelay(2000);

		//ベルトコンベアーを一定速度で進める Aの出番はなさそう
		Motor.B.setSpeed(25);
		Motor.B.forward();
		Motor.A.setSpeed(250);
		Motor.A.rotate(360);
		
		boolean gotAllGoods = false;
		byte goodsCount = 0;
		byte weightColor = -1;
		byte valueColor = -1;
		int getEnvironmentCount = 0;

		while(Button.ESCAPE.isUp() && !gotAllGoods)	{

			Delay.msDelay(20);

			env = getEnvironment();
			getEnvironmentCount++;

			
			if(env != -1){
				
				byte prevEnv = env;
				Delay.msDelay(50);
				env = getEnvironment();
				getEnvironmentCount++;

				// 誤認識でないか再度色を確かめて、問題なければ読み取った色を保存する
				if(prevEnv == env){

					if(weightColor == -1){
						// 重さの色が未決定なら格納
						weightColor = env;
					}

					else if(env != weightColor && valueColor == -1){
						// 価値の色が未決定なら格納
						// ただし、重さの色と異なる場合のみ
						// 重さの色と価値の色が同じ場合は、荷物の読み取り終了時に処理する
						valueColor = env;
						Motor.A.rotate(360);

					}
				}
			}

			else { // 環境色(色なし・黒・緑)と判定された場合

				if(goodsCount < 1 && weightColor == -1){
					//まだ荷物の始点に到達していない 
					getEnvironmentCount = 0; 
				}
				
				if(weightColor != -1){
					// 1つの荷物の終点に達していると思われ、
					// 少なくとも重さの色は決定しているはずである
					// 誤認識でないか再度色を確かめて大丈夫なら
					// 荷物1つの読み取りを終了したと認識する
					byte prevEnv = env;
					Delay.msDelay(50);
					env = getEnvironment();
					getEnvironmentCount++;
					
					if(prevEnv == env){

						if(valueColor == -1){
							// 重さの色と価値の色が同じ場合だと思われるので、そう処理する
							valueColor = weightColor;
						}

						// ここで重さと価値の色のペアが決定しているはずなので、後処理をする
						goodsWeight[goodsCount] = weightColor2Weight[weightColor];
						goodsValue[goodsCount] = valueColor2Value[valueColor];
						goodsCount++;
						if(goodsCount == NUMBER_GOODS){
							gotAllGoods = true;
						}
						else{
							weightColor = -1;
							valueColor = -1;
						}

					}

				}

			}
			
		writeMess("color & count", "wColor & vColor", env, goodsCount, weightColor, valueColor);	
			
		}

		/* STEP 2
		 *  荷物を最初の状態まで巻き戻す
		 */

		
		Motor.B.stop();

		Motor.B.setSpeed(100);
		Motor.B.forward();
		Delay.msDelay(3000);
		
		Motor.B.stop();

		Delay.msDelay(1000);
		Motor.B.setSpeed(30 * 5); // 5倍速で巻き戻す
		Motor.B.backward();
		Delay.msDelay( (int) (2*getEnvironmentCount + 500) );
	
		Motor.B.stop();

		Motor.D.setSpeed(100);
		Motor.D.rotate(30);

		/* STEP 2.5
		 * ナップザック問題を解く
		 * 遺伝的アルゴリズムを使用する場合は
		 * http://qiita.com/simanezumi1989/items/10cfe1e8a23cd9d4c7b1
		 * を参考にする
		 * 【現在は総当たり】
		 */

		// E[i]:i番目の荷物を選ぶかどうか（0,1で決める）
		byte E[] = new byte[NUMBER_GOODS];
		for (int i=0;i<NUMBER_GOODS;i++){
			E[i] = 0; // 初期値は「全部選ばない」
		}

		// 総当たりで価値最大になったパターン
		int maxValue = 0;

		// 現状最善のEを保存
		byte F[] = new byte[NUMBER_GOODS];
		for (int i=0;i<NUMBER_GOODS;i++){
			F[i] = 0; // 初期値は「全部選ばない」
		}

		// 総当たりで評価
		for (int caseCount = 0; caseCount < NUMBER_GOODS * NUMBER_GOODS; caseCount++){

			// 総価値を計算
			int weightSum = 0;
			int valueSum = 0;
			for (int i=0;i<NUMBER_GOODS;i++) {
				// 1ならナップザックに入れる
				if(E[i] == 1){
					weightSum += goodsWeight[i];
					valueSum += goodsValue[i];
				}
				// 容量オーバーは価値0
				if(weightSum > WEIGHT_LIMIT){
					valueSum = 0;
				}
			}

			// 総価値の最大値を更新
			if(valueSum > maxValue){
				maxValue = valueSum;
				for (int i=0;i<NUMBER_GOODS;i++) {
					F[i] = E[i];
				}
			}

			// 次状態の準備
			// 次のcaseCountの値を2進数の0/1で表したものを逆順でそのままあてはめる
			int tmp = caseCount + 1;
			if(tmp != NUMBER_GOODS * NUMBER_GOODS){
				byte idx = 0;
				while(idx<NUMBER_GOODS){
					if(tmp % 2 == 0){
						E[idx] = 0;
						tmp /= 2;
					}
					else{
						E[idx] = 1;
						tmp--;
						tmp /= 2;
					}
					idx++;
				}
			}

		}

		/* STEP 3
		 * ナップザック問題を解いた結果に基づいて、必要な荷物を残し、他をはじく
		 */

		boolean phasechanger = false;
		Motor.B.setSpeed(100);
		Motor.B.forward();
		while(Button.ESCAPE.isUp() && !phasechanger)	{
			Delay.msDelay(10);
			env = getEnvironment();
			if(env != -1){
				phasechanger = true;
				Motor.B.stop();
			}
		}
		
		// 最初の荷物がアーム(ハンマー)のところに到達するまでの待機時間
		Motor.B.setSpeed(103);
		Motor.B.forward();
		Delay.msDelay(3300); // 【要調整】
		Motor.B.stop();

		for (int i=0;i<NUMBER_GOODS;i++){
			// 不要な荷物ならはじく
			if(F[i] == 0){
				Motor.D.setSpeed(300);
				Motor.D.rotate(-30);
				Motor.D.rotate(30);
			}

			// 次の荷物の位置まで進む
			Motor.B.setSpeed(54);
			Motor.B.forward();
			Delay.msDelay(1000); //【要調整】
			Motor.B.stop();
		}

		// 残った荷物が流れ落ちるまで十分な時間コンベアを回す
		Motor.B.setSpeed(500);
		Motor.B.forward();
		Delay.msDelay(2000);

		exit();
		
	}


	/* Void instruction for closing the program safely
	 * Also closing the used sensors.
	 */
	public static void exit(){ // Please insert the amount of sensors used
		Motor.B.stop();
		Motor.D.rotate(-30);
		pilot.stop();
		sensor3.close();
		System.exit(0);

	}

	// Functions that Stores actions the robot can take.
	public static void puppyCommands(byte command){

		switch(command){

		case 0: //drop <- red
			Motor.B.setSpeed(90);
			Motor.B.forward();
			Delay.msDelay(2500);
			Motor.B.stop();

			Motor.D.setSpeed(300);
			Motor.D.rotate(-60);
			Delay.msDelay(2000);
			Motor.D.rotate(60);

			break;

		case 1: //pass <- other color
			Motor.B.setSpeed(90);
			Motor.B.forward();
			Delay.msDelay(2500);
			Motor.B.stop();

			//rotate reverse ways to debug
			Motor.D.setSpeed(300);
			Motor.D.rotate(60);
			Delay.msDelay(2000);
			Motor.D.rotate(-60);

			break;

		}

	}

	public static byte getEnvironment(){

		byte colorSensor = color.pressed();
		// 0 : RED
		// 2 : BLUE
		// 3 : YELLOW
		// -1: others -> think as no color and continue to move the belt
		return colorSensor;

	}

	public static void writeMess(String mess1, String mess2, int val0, int val1, int val2, int val3){

		LCD.clear();

		LCD.drawString(mess1, 2, 1); LCD.drawInt(val0, 4, 2); LCD.drawInt(val1, 12, 2);
		LCD.drawString(mess2, 2, 3); LCD.drawInt(val2, 4, 4); LCD.drawInt(val3, 12, 4);

//		LCD.drawString(mess2, 2, 4); LCD.drawInt((int)(val3*100), 11, 4);

		//LCD.drawString("L.R = ", 1, 6); LCD.drawInt((int)(QTools.LEARN_RATE*100), 6, 6);
//		LCD.drawString("E.R = ", 10, 6); LCD.drawInt((int)(EXPLORE_RATE*100), 15, 6);

	}
	
	
///////////////////////////TIMER
	
	public class Timer {
	    public static final String MILLISECOND = "ms";
	    public static final String SECOND = "s";
	    public static final String NANOSECOND = "ns";
	     
	    private long t0, time;
	    private String timeUnit;
	     
	    public Timer() {
	        this(NANOSECOND);
	    }
	     
	    public Timer(String timeUnit) {
	        setTimeUnit(timeUnit);
	    }
	     
	    public String getTimeUnit() {
	        return timeUnit;
	    }
	     
	    public final void setTimeUnit(String timeUnit) {
	        this.timeUnit = timeUnit;
	    }
	     
	    public void start() {
	        t0 = System.nanoTime();
	    }
	     
	    public void stop() {
	        time = System.nanoTime() - t0;
	    }
	     
	    public double getTime() {
	        if(timeUnit.equals(NANOSECOND)) return time;
	        if(timeUnit.equals(SECOND)) return time / 1000000000.0;
	        if(timeUnit.equals(MILLISECOND)) return time / 1000000.0;
	        return time;
	    }
	}

	
	
}
