package Pack;

import lejos.hardware.lcd.LCD;
//import lejos.utility.Delay;

public class QTools {
	
	/* Values involved in the robot's learning process.
	 */
	
	public static float LEARN_RATE = 0.35f; // Value involved in the stimulation of good actions
	//public static float EXPLORE_RATE = 0.15f; // Value involved on the exploration of new actions
	
	byte ls = -1; // Last state
	byte la = -1; // Last action
	byte MaxA = 0; // Action with the highest q-value
	
	float[][] Q; // Q-table values
	
	float r = 0; // Last reward value
	
	byte[] state;
	byte[] action;
		
	
	/* Constructor for the Q values
	 * matrix.
	 */
	public QTools(byte actions, byte states){
		//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
		state = new byte[states]; // The number 1 means the amount of sensors you'll read
		action = new byte[actions];
		initAction(action);
		initState(state);
		Qinit(actions, states);
	}
	
	/* Initialize the Q table with random variables
	 * These values will change while the program is running.
	 */
	public void Qinit(byte actions, byte states){
		
		Q = new float[actions][states];
		setRandomVal(Q);
		
	}
	
	/* Get the next action the robot will apply
	 * Also Apply the Learning Reinforcement Algorithm.
	 */
	
	//States == env == ŏIIɊUFƂ̔ԍ
	public byte getAction(byte states){
		
		writeMess();
		
		byte st = getState(states); //Environment -> State ID (New state)
		
		if(ls >= 0 || la >= 0){
			
			r = reward(ls, la);
			
			
			if(r == 1){ // In case that the reward is equal to one				
				
				if(LEARN_RATE > 0){
					LEARN_RATE = LEARN_RATE - 0.001f;
				}
				
			}
			
			// Application of Reinforcement Learning algorithm
			MaxA = getMaxAction(st);   //**********************************************************
			Q[la][ls] = Q[la][ls] + LEARN_RATE*(r + Q[MaxA][st] - Q[la][ls]);
			
		}
		
		ls = st;
		
		la = boltzProb(ls);
		
		/*if((float) Math.random() > EXPLORE_RATE){
			la = MaxA;
		}else{
			la = (byte)(Math.random()*4); // 4 is the number of actions
		}*/
		
		return la;
		
	}
	
	/* Function that gets the Maximum value from the Q table
	 * Inside of the specific state
	 */	
	public byte getMaxAction(byte st){   //****************************************
		
		float max = -1000;
		byte action = 0;
		
		for(byte i = 0; i < Q.length; ++i){  //*********************************************
			
			if(Q[i][st] > max){
				max = Q[i][st];
				action = i;
			}
			
		}
		return action;
		
	}
	
	
	/* Initialize the Q-values matrix with random values.
	 * The size of this matrix depends on the amount of states and actions.
	 */
	public void setRandomVal(float[][] Q){
		for(byte i=0; i<Q.length; ++i){
			for(byte j=0; j<Q[0].length; ++j){
				Q[i][j] = (float)Math.random();
				//                                        System.out.println("random = " + Q[i][j]);
			}
		}
	}
	
	/* Initialize the matrix of states given the amount of
	 * combinations of the sensors (That establish the environment)
	 * In this robots exists 2 contact sensors with ON/OFF states
	 * So, it will give 2x2 combinations.
	 */
	public static void initState(byte[] state){
	
		for(byte i=0; i<state.length; ++i){  // In this case it's just an array of 4 states.
			state[i] = i;         // Because we just read one sensor
		}
	}
	
	/*Initialize the matrix of actions the robot can make
	 * regarding both motors as one entity.
	 */
	public static void initAction(byte[] action){
		
		for(byte i=0; i<action.length; ++i){
			action[i] = i;
		}
	}

	/*Giving an action ID (a) from 1~theMaxNumberOfActions, give back
	 * the action related with the ID.
	 */
	public byte getCommand(byte a){
		return action[a];
	}
	
	public float reward(byte s, byte a){
		float r = 0;
		switch(s){
		case 0:  // RED
			switch(a){
			case 0:  // right
				r = 1;  // reward
				break;
				
			case 1:  // left
				r = -1f;  // punishment
				break;
			case 2: //pass
				r = -1f;
				break;
								
			default: break;
			}
			
			break;
				
		case 2:  // BLUE
			
			switch(a){
			case 0:  //right 
				r = -1f;  // punishment
				break;
				
			case 1:  // left
				r = 1;  // reward
				break;
			case 2: //pass
				r = -1f;
				break;
			default: break;
			}

			break;
			
		case 3:  // YELLOW
			switch(a){
			case 0:  // right
				r = -1f;  // punishment
				break;
				
			case 1:  // left
				r = -1f;  // reward
				break;
				
			case 2: //pass
				r = 1;
				break;
			default: break;
			}
			
		default: break;
		}
		
		return r;
		
	}
	
	/* Given an array of environments or actual states of the sensors
	 * Compare with the states matrix and get the ID of the matching one.
	 */
	
	
	public byte getState(byte states){
		
		for(byte i=0; i<state.length; ++i){
			
			if(states == state[i]){
				return states;
			}
			
		}
		return -1;
	}
	
	public void writeMess(){
		
		//String mess1, String mess2, int val1, int val2, float val3
		
		//LCD.clear();
		//LCD.drawString(mess1, 2, 1); LCD.drawInt(val1, 4, 2); LCD.drawInt(val2, 12, 2);
		
		//LCD.drawString(mess2, 2, 4); LCD.drawInt((int)(val3*100), 11, 4);
		
		LCD.drawString("L.R = ", 1, 6); LCD.drawInt((int)(LEARN_RATE*100), 6, 6);
		//LCD.drawString("E.R = ", 10, 6); LCD.drawInt((int)(EXPLORE_RATE*100), 15, 6);
		
	}
	
	public byte boltzProb(byte state){
		
		//Boltzmann probability value.
		float p = 0, max = -100;
		byte a = -1;
		byte max1 = getMaxAction(state);
		
		for(byte i = 0; i < this.action.length; i++){
			p = (float)(Math.exp(Q[i][state] - Q[max1][state])/sumQvals(state));
			
			if(p > max){
				max = p;
				a = i;
			}
			
		}
		return a;
		
	}
	
	public float sumQvals(byte state){
		
		float sum = 0;
		byte max = getMaxAction(state);

		for(byte i = 0; i < this.action.length; i++){
			sum = sum + (float)Math.exp(Q[i][state] - Q[max][state]);
		}
		
		return sum;
		
	}
	
}
